@@ -108,10 +108,6 @@ class Selenese(object):
     def open(self, url, ignoreResponseCode=True):
         self.selenium.do_command("open", [url, ignoreResponseCode])
 
-    @passthrough
-    def selectWindow(self, name):
-        pass
-
     @passthrough
     def addCustomRequestHeader(self, key, value):
         pass
@@ -305,6 +301,12 @@ class Selenese(object):
     def mouseUpRightAt(self, locator, coord):
         pass
 
+    def openWindow(self, url, window_id):
+        if window_id == 'null':
+            raise ValueError("Cannot name a window 'null' "
+                             "as this name is used be Selenium internally.")
+        return self.selenium.open_window(url, window_id)
+
     def refresh(self):
         # No thanks to selenium... why would one ever *not* want to wait for
         # the page to load?
@@ -335,6 +337,9 @@ class Selenese(object):
     def selectFrame(self):
         pass
 
+    def selectWindow(self, window_id=None):
+        return self.selenium.select_window(window_id or 'null')
+
     @passthrough
     def submit(self, locator):
         pass
@@ -410,14 +415,17 @@ class Selenese(object):
                 'No alert occured.')
         return self.selenium.get_alert()
 
+    @assert_type('list')
     @passthrough
     def getAllWindowIds(self):
         pass
 
-    @passthrough
+    @assert_type('list')
     def getAllWindowNames(self):
-        pass
+        return [name for name in self.selenium.get_all_window_names()
+                if name != 'null']
 
+    @assert_type('list')
     @passthrough
     def getAllWindowTitles(self):
         pass
@@ -685,6 +693,9 @@ class Selenese(object):
         elif getter.assert_type == 'locator_pattern':
             return lambda locator, pattern: self._assert_pattern(
                 getter, requested_name, pattern, locator)
+        elif getter.assert_type == 'list':
+            return lambda expected: self._assert_list(
+                getter, requested_name, expected)
         elif getter.assert_type is None:
             return lambda: self._assert(getter, requested_name)
         else:
@@ -705,6 +716,31 @@ class Selenese(object):
                 'Expected: %r, got: %r from %s' %
                 (pattern, result, self._call_repr(name, *args)))
 
+    def _assert_list(self, getter, name, expected):
+        result = getter()
+        if expected != result:
+            detail = ''
+            if len(expected) != len(result):
+                detail += ('Expected %s items, got %s items.\n' %
+                           (len(expected), len(result)))
+            if len(expected) < len(result):
+                detail += ('First extra element: %r\n\n' %
+                           (result[len(expected)],))
+            elif len(expected) > len(result):
+                detail += ('First missing element: %r\n\n' %
+                           (expected[len(result)],))
+            else:
+                for i, x in enumerate(expected):
+                    if x != result[i]:
+                        detail += (
+                            'First differing list item at index %s:\n'
+                            '- %r\n+ %r\n\n' % (i, x, result[i]))
+                        break
+            raise self.failureException(
+                detail + ('Expected: %s,\ngot: %s\nfrom %s' % (
+                        abbrev_repr(expected), abbrev_repr(result),
+                        self._call_repr(name))))
+
     def _negate(self, assertion, name, *args, **kw):
         try:
             assertion(*args, **kw)
@@ -778,3 +814,10 @@ def selenese_pattern_equals(text, pattern):
         except KeyError:
             pass
     return matcher(text, pattern)
+
+
+def abbrev_repr(x, size=70):
+    r = repr(x)
+    if len(r) > size:
+        r = r[:size-3] + '...'
+    return r
